from functools import partial
import numpy as np
import scipy
from typing import Optional, Callable

from meanfi.params.rparams import rparams_to_tb, tb_to_rparams
from meanfi.tb.tb import add_tb, _tb_type
from meanfi.mf import density_matrix, meanfield
from meanfi.model import Model
from meanfi.tb.utils import fermi_energy


def cost_mf(mf_param: np.ndarray, model: Model, nk: int = 20) -> np.ndarray:
    """Defines the cost function for root solver.

    The cost function is the difference between the computed and inputted mean-field.

    Parameters
    ----------
    mf_param :
        1D real array that parametrises the mean-field correction.
    Model :
        Interacting tight-binding problem definition.
    nk :
        Number of k-points in a grid to sample the Brillouin zone along each dimension.
        If the system is 0-dimensional (finite), this parameter is ignored.

    Returns
    -------
    :
        1D real array that is the difference between the computed and inputted mean-field
        parametrisations
    """
    shape = model._ndof
    mf = rparams_to_tb(mf_param, list(model.h_int), shape)
    mf_new = model.mfield(mf, nk=nk)
    mf_params_new = tb_to_rparams(mf_new)
    return mf_params_new - mf_param


def cost_density(rho_params: np.ndarray, model: Model, nk: int = 20) -> np.ndarray:
    """Defines the cost function for root solver.

    The cost function is the difference between the computed and inputted density matrix
    reduced to the hoppings only present in the h_int.

    Parameters
    ----------
    rho_params :
        1D real array that parametrises the density matrix reduced to the
        hoppings (keys) present in h_int.
    Model :
        Interacting tight-binding problem definition.
    nk :
        Number of k-points in a grid to sample the Brillouin zone along each dimension.
        If the system is 0-dimensional (finite), this parameter is ignored.

    Returns
    -------
    :
        1D real array that is the difference between the computed and inputted
        density matrix parametrisations reduced to the hoppings present in h_int.
    """
    shape = model._ndof
    rho_reduced = rparams_to_tb(rho_params, list(model.h_int), shape)
    rho_new = model.density_matrix(rho_reduced, nk=nk)
    rho_reduced_new = {key: rho_new[key] for key in model.h_int}
    rho_params_new = tb_to_rparams(rho_reduced_new)
    return rho_params_new - rho_params


def solver_mf(
    model: Model,
    mf_guess: np.ndarray,
    nk: int = 20,
    optimizer: Optional[Callable] = scipy.optimize.anderson,
    optimizer_kwargs: Optional[dict[str, str]] = {"M": 0},
) -> _tb_type:
    """Solve for the mean-field correction through self-consistent root finding
    by finding the mean-field correction fixed point.

    Parameters
    ----------
    model :
        Interacting tight-binding problem definition.
    mf_guess :
        The initial guess for the mean-field correction in the tight-binding dictionary format.
    nk :
        Number of k-points in a grid to sample the Brillouin zone along each dimension.
        If the system is 0-dimensional (finite), this parameter is ignored.
    optimizer :
        The solver used to solve the fixed point iteration.
        Default uses `scipy.optimize.anderson`.
    optimizer_kwargs :
        The keyword arguments to pass to the optimizer.

    Returns
    -------
    :
        Mean-field correction solution in the tight-binding dictionary format.
    """
    shape = model._ndof
    mf_params = tb_to_rparams(mf_guess)
    f = partial(cost_mf, model=model, nk=nk)
    result = rparams_to_tb(
        optimizer(f, mf_params, **optimizer_kwargs), list(model.h_int), shape
    )
    fermi = fermi_energy(add_tb(model.h_0, result), model.filling, nk=nk)
    return add_tb(result, {model._local_key: -fermi * np.eye(model._ndof)})


def solver_density(
    model: Model,
    mf_guess: _tb_type,
    nk: int = 20,
    optimizer: Optional[Callable] = scipy.optimize.anderson,
    optimizer_kwargs: Optional[dict[str, str]] = {"M": 0, "line_search": "wolfe"},
) -> _tb_type:
    """Solve for the mean-field correction through self-consistent root finding
    by finding the density matrix fixed point.

    Parameters
    ----------
    model :
        Interacting tight-binding problem definition.
    mf_guess :
        The initial guess for the mean-field correction in the tight-binding dictionary format.
    nk :
        Number of k-points in a grid to sample the Brillouin zone along each dimension.
        If the system is 0-dimensional (finite), this parameter is ignored.
    optimizer :
        The solver used to solve the fixed point iteration.
        Default uses `scipy.optimize.anderson`.
    optimizer_kwargs :
        The keyword arguments to pass to the optimizer.

    Returns
    -------
    :
        Mean-field correction solution in the tight-binding dictionary format.
    """
    shape = model._ndof
    rho_guess = density_matrix(
        add_tb(model.h_0, mf_guess), filling=model.filling, nk=nk
    )[0]
    rho_guess_reduced = {key: rho_guess[key] for key in model.h_int}

    rho_params = tb_to_rparams(rho_guess_reduced)
    f = partial(cost_density, model=model, nk=nk)
    rho_result = rparams_to_tb(
        optimizer(f, rho_params, **optimizer_kwargs), list(model.h_int), shape
    )
    mf_result = meanfield(rho_result, model.h_int)
    fermi = fermi_energy(add_tb(model.h_0, mf_result), model.filling, nk=nk)
    return add_tb(mf_result, {model._local_key: -fermi * np.eye(model._ndof)})


solver = solver_density



# %%
import pytest
import numpy as np
from meanfi.params.rparams import rparams_to_tb, tb_to_rparams
from meanfi.tb.tb import compare_dicts
from meanfi.tb.utils import generate_tb_keys
from meanfi import guess_tb

repeat_number = 10
# %%
ndof = 10
vectors = ((0, 0), (1, 0), (-1, 0), (0, 1), (0, -1), (1, -1), (-1, 1), (1, 1), (-1, -1))


@pytest.mark.parametrize("seed", range(repeat_number))
def test_parametrisation(seed):
    """Test the parametrisation of the tight-binding model."""
    np.random.seed(seed)
    for ndim in np.arange(4):
        cutoff = np.random.randint(0, 4)
        ndof = np.random.randint(1, 10)

        keys = generate_tb_keys(cutoff, ndim)
        mf_guess = guess_tb(keys, ndof)

        mf_params = tb_to_rparams(mf_guess)
        mf_new = rparams_to_tb(mf_params, list(mf_guess), ndof)
        compare_dicts(mf_guess, mf_new)



import numpy as np

from meanfi.mf import (
    density_matrix,
    meanfield,
)
from meanfi.tb.tb import add_tb, _tb_type


def _check_hermiticity(h):
    for vector in h.keys():
        op_vector = tuple(-1 * np.array(vector))
        op_vector = tuple(-1 * np.array(vector))
        if not np.allclose(h[vector], h[op_vector].conj().T):
            raise ValueError("Tight-binding dictionary must be hermitian.")


def _tb_type_check(tb):
    for count, key in enumerate(tb):
        if not isinstance(tb[key], np.ndarray):
            raise ValueError(
                "Values of the tight-binding dictionary must be numpy arrays"
            )
        shape = tb[key].shape
        if count == 0:
            size = shape[0]
        if not len(shape) == 2:
            raise ValueError(
                "Values of the tight-binding dictionary must be square matrices"
            )
        if not size == shape[0]:
            raise ValueError(
                "Values of the tight-binding dictionary must have consistent shape"
            )


class Model:
    """
    Data class which defines the interacting tight-binding problem.

    Parameters
    ----------
    h_0 :
        Non-interacting hermitian Hamiltonian tight-binding dictionary.
    h_int :
        Interaction hermitian Hamiltonian tight-binding dictionary.
    filling :
        Number of particles in a unit cell.
        Used to determine the Fermi level.

    Notes
    -----

    The interaction h_int must be of density-density type.
    For example, h_int[(1,)][i, j] = V means a repulsive interaction
    of strength V between two particles with internal degrees of freedom i and j
    separated by 1 lattice vector.
    """

    def __init__(self, h_0: _tb_type, h_int: _tb_type, filling: float) -> None:
        _tb_type_check(h_0)
        self.h_0 = h_0
        _tb_type_check(h_int)
        self.h_int = h_int
        if not isinstance(filling, (float, int)):
            raise ValueError("Filling must be a float or an integer")
        if not filling > 0:
            raise ValueError("Filling must be a positive value")
        self.filling = filling

        _first_key = list(h_0)[0]
        self._ndim = len(_first_key)
        self._ndof = h_0[_first_key].shape[0]
        self._local_key = tuple(np.zeros((self._ndim,), dtype=int))

        _check_hermiticity(h_0)
        _check_hermiticity(h_int)

    def density_matrix(self, rho: _tb_type, nk: int = 20) -> _tb_type:
        """Computes the density matrix from a given initial density matrix.

        Parameters
        ----------
        rho :
            Initial density matrix tight-binding dictionary.
        nk :
            Number of k-points in a grid to sample the Brillouin zone along each dimension.
            If the system is 0-dimensional (finite), this parameter is ignored.

        Returns
        -------
        :
            Density matrix tight-binding dictionary.
        """
        mf = meanfield(rho, self.h_int)
        return density_matrix(add_tb(self.h_0, mf), self.filling, nk)[0]

    def mfield(self, mf: _tb_type, nk: int = 20) -> _tb_type:
        """Computes a new mean-field correction from a given one.

        Parameters
        ----------
        mf :
            Initial mean-field correction tight-binding dictionary.
        nk :
            Number of k-points in a grid to sample the Brillouin zone along each dimension.
            If the system is 0-dimensional (finite), this parameter is ignored.

        Returns
        -------
        :
            new mean-field correction tight-binding dictionary.
        """
        rho, fermi_energy = density_matrix(add_tb(self.h_0, mf), self.filling, nk)
        return add_tb(
            meanfield(rho, self.h_int),
            {self._local_key: -fermi_energy * np.eye(self._ndof)},
        )



# %%
import numpy as np
import pytest

from meanfi.tests.test_graphene import compute_gap
from meanfi import (
    Model,
    fermi_energy,
    solver,
    guess_tb,
    add_tb,
)

repeat_number = 3


# %%
def gap_relation_hubbard(Us, nk, nk_dense, tol=1e-3):
    """Gap relation for the Hubbard model.

    Parameters
    ----------
    Us : np.array
        The Hubbard U parameter.
    nk : int
        The number of k-points to use in the grid.
    nk_dense : int
        The number of k-points to use in the dense grid for calculating the gap.
    tol : float
        The tolerance for the fitting of the gap.
    """
    hopp = np.kron(np.array([[0, 1], [0, 0]]), np.eye(2))
    h_0 = {(0,): hopp + hopp.T.conj(), (1,): hopp, (-1,): hopp.T.conj()}
    gaps = []
    for U in Us:
        h_int = {
            (0,): U * np.kron(np.eye(2), np.ones((2, 2))),
        }
        guess = guess_tb(frozenset(h_int), len(list(h_0.values())[0]))
        full_model = Model(h_0, h_int, filling=2)
        mf_sol = solver(full_model, guess, nk=nk)
        mf_full = add_tb(h_0, mf_sol)
        EF = fermi_energy(mf_full, filling=2, nk=nk)
        _gap = compute_gap(mf_full, fermi_energy=EF, nk=nk_dense)
        gaps.append(_gap)

    fit_gap = np.polyfit(Us, np.array(gaps), 1)[0]
    assert np.abs(fit_gap - 1) < tol


@pytest.mark.parametrize("seed", range(repeat_number))
def test_gap_hubbard(seed):
    """Test the gap prediction for the Hubbard model."""
    np.random.seed(seed)
    Us = np.linspace(8, 10, 15, endpoint=True)
    gap_relation_hubbard(Us, nk=20, nk_dense=100, tol=1e-1)



# %%
import numpy as np
import pytest

from meanfi import (
    Model,
    solver,
    guess_tb,
    scale_tb,
    add_tb,
    expectation_value,
    density_matrix,
)

from meanfi.tb.utils import generate_tb_keys


# %%
def total_energy(ham_tb, rho_tb):
    return np.real(expectation_value(rho_tb, ham_tb))


# %%
U0 = 1
filling = 2
nk = 10
repeat_number = 3
ndof = 4
cutoff = 1


# %%
@np.vectorize
def mf_rescaled(alpha, mf0, h0):
    hamiltonian = add_tb(h0, scale_tb(mf0, alpha))
    rho, _ = density_matrix(hamiltonian, filling=filling, nk=nk)
    hamiltonian = add_tb(h0, scale_tb(mf0, np.sign(alpha)))
    return total_energy(hamiltonian, rho)


@pytest.mark.parametrize("seed", range(repeat_number))
def test_mexican_hat(seed):
    np.random.seed(seed)
    h0s = []
    h_ints = []
    for ndim in np.arange(4):
        keys = generate_tb_keys(cutoff, ndim)
        h0s.append(guess_tb(keys, ndof))
        h_int = guess_tb(keys, ndof)
        h_int[keys[len(keys) // 2]] += U0
        h_ints.append(h_int)

    for h0, h_int in zip(h0s, h_ints):
        guess = guess_tb(frozenset(h_int), ndof)
        _model = Model(h0, h_int, filling=filling)
        mf_sol_groundstate = solver(
            _model, mf_guess=guess, nk=nk, optimizer_kwargs={"M": 0}
        )

        alphas = np.random.uniform(0, 50, 100)
        alphas = np.where(alphas == 1, 0, alphas)
        assert np.all(
            mf_rescaled(alphas, mf0=mf_sol_groundstate, h0=h0)
            > mf_rescaled(np.array([1]), mf0=mf_sol_groundstate, h0=h0)
        )



import numpy as np

from meanfi.tb.tb import _tb_type


def complex_to_real(z: np.ndarray) -> np.ndarray:
    """Split and concatenate real and imaginary parts of a complex array.

    Parameters
    ----------
    z :
        Complex array.

    Returns
    -------
    :
        Real array that concatenates the real and imaginary parts of the input array.
    """
    return np.concatenate((np.real(z), np.imag(z)))


def real_to_complex(z: np.ndarray) -> np.ndarray:
    """Undo `complex_to_real`."""
    return z[: len(z) // 2] + 1j * z[len(z) // 2 :]


def tb_to_rparams(tb: _tb_type) -> np.ndarray:
    """Parametrise a hermitian tight-binding dictionary by a real vector.

    Parameters
    ----------
    tb :
        tight-binding dictionary.

    Returns
    -------
    :
        1D real vector that parametrises the tight-binding dictionary.
    """
    ndim = len(list(tb)[0])
    onsite_key = tuple(np.zeros((ndim,), dtype=int))

    hopping_keys = sorted([key for key in tb.keys() if key != onsite_key])
    hopping_keys = hopping_keys[: len(hopping_keys) // 2]

    onsite_upper_vals = tb[onsite_key][
        np.triu_indices(tb[onsite_key].shape[-1], k=1)
    ].flatten()
    onsite_diag_vals = np.diag(tb[onsite_key]).real.flatten()

    hopping_vals = [tb[key].flatten() for key in hopping_keys]
    complex_vals = np.concatenate([onsite_upper_vals, *hopping_vals])
    return np.concatenate([onsite_diag_vals, complex_to_real(complex_vals)])


def rparams_to_tb(
    tb_params: np.ndarray, tb_keys: list[tuple[None] | tuple[int, ...]], ndof: int
) -> _tb_type:
    """Extract a hermitian tight-binding dictionary from a real vector parametrisation.

    Parameters
    ----------
    tb_params :
        1D real array that parametrises the tight-binding dictionary.
    tb_keys :
        List of keys of the tight-binding dictionary.
    ndof :
        Number internal degrees of freedom within the unit cell.

    Returns
    -------
    :
        Tight-biding dictionary.
    """
    ndim = len(tb_keys[0])

    hopping_shape = (len(tb_keys) - 1, ndof, ndof)

    onsite_idxs = ndof + ndof * (ndof - 1) // 2
    onsite_key = tuple(np.zeros((ndim,), dtype=int))

    # reconstruct the complex values
    onsite_diag_vals = tb_params[:ndof]
    complex_vals = real_to_complex(tb_params[ndof:])
    onsite_upper_vals = complex_vals[: onsite_idxs - ndof]
    hopping_vals = complex_vals[(onsite_idxs - ndof) :]

    # first build onsite matrix
    onsite_matrix = np.zeros((ndof, ndof), dtype=complex)
    onsite_matrix[np.triu_indices(ndof, k=1)] = onsite_upper_vals
    onsite_matrix += onsite_matrix.conj().T
    onsite_matrix[np.diag_indices(ndof)] = onsite_diag_vals

    # then build hopping matrices
    hopping_matrices = np.zeros(hopping_shape, dtype=complex)
    N = len(tb_keys) // 2
    hopping_matrices[:N] = hopping_vals.reshape(N, *hopping_shape[1:])
    hopping_matrices[N:] = np.moveaxis(
        np.flip(hopping_matrices[:N], axis=0), -1, -2
    ).conj()

    # combine all into a dictionary
    hopping_keys = sorted([key for key in tb_keys if key != onsite_key])
    tb = {key: hopping_matrices[i] for i, key in enumerate(hopping_keys)}
    tb[onsite_key] = onsite_matrix
    return tb



import kwant
import numpy as np

from meanfi.kwant_helper.utils import build_interacting_syst

s0 = np.identity(2)
sz = np.diag([1, -1])


def graphene_extended_hubbard():
    """Create a graphene system with extended Hubbard interactions.

    Return:
    ------
    bulk_graphene : kwant.builder.Builder
        The bulk graphene system.
    syst_V : kwant.builder.Builder
    """
    # Create graphene lattice
    graphene = kwant.lattice.general(
        [[1, 0], [1 / 2, np.sqrt(3) / 2]], [[0, 0], [0, 1 / np.sqrt(3)]], norbs=2
    )
    a, b = graphene.sublattices

    # Create bulk system
    bulk_graphene = kwant.Builder(kwant.TranslationalSymmetry(*graphene.prim_vecs))
    # Set onsite energy to zero
    bulk_graphene[a.shape((lambda pos: True), (0, 0))] = 0 * s0
    bulk_graphene[b.shape((lambda pos: True), (0, 0))] = 0 * s0
    # Add hoppings between sublattices
    bulk_graphene[graphene.neighbors(1)] = s0

    def onsite_int(site, U):
        return U * np.ones((2, 2))

    def nn_int(site1, site2, V):
        return V * np.ones((2, 2))

    syst_V = build_interacting_syst(
        builder=bulk_graphene,
        lattice=graphene,
        func_onsite=onsite_int,
        func_hop=nn_int,
    )

    return bulk_graphene, syst_V



import itertools
import numpy as np
from typing import Callable
from scipy.fftpack import ifftn

from meanfi.tb.tb import _tb_type


def tb_to_kgrid(tb: _tb_type, nk: int) -> np.ndarray:
    """Evaluate a tight-binding dictionary on a k-space grid.

    Parameters
    ----------
    tb :
        Tight-binding dictionary to evaluate on a k-space grid.
    nk :
        Number of k-points in a grid to sample the Brillouin zone along each dimension.
        If the system is 0-dimensional (finite), this parameter is ignored.

    Returns
    -------
    :
        Tight-binding dictionary evaluated on a k-space grid.
        Has shape (nk, nk, ..., ndof, ndof), where ndof is number of internal degrees of freedom.
    """
    ndim = len(list(tb)[0])
    ks = np.linspace(-np.pi, np.pi, nk, endpoint=False)
    ks = np.concatenate((ks[nk // 2 :], ks[: nk // 2]), axis=0)  # shift for ifft
    kgrid = np.meshgrid(*([ks] * ndim), indexing="ij")

    num_keys = len(list(tb.keys()))
    tb_array = np.array(list(tb.values()))
    keys = np.array(list(tb.keys()))

    k_dependency = np.exp(-1j * np.tensordot(keys, kgrid, 1))[
        (...,) + (np.newaxis,) * 2
    ]
    tb_array = tb_array.reshape([num_keys] + [1] * ndim + list(tb_array.shape[1:]))
    return np.sum(tb_array * k_dependency, axis=0)


def kgrid_to_tb(kgrid_array: np.ndarray) -> _tb_type:
    """
    Convert a k-space grid array to a tight-binding dictionary.

    Parameters
    ----------
    kgrid_array :
        K-space grid array to convert to a tight-binding dictionary.
        The array should be of shape (nk, nk, ..., ndof, ndof),
        where ndof is number of internal degrees of freedom.
    Returns
    -------
    :
        Tight-binding dictionary.
    """
    ndim = len(kgrid_array.shape) - 2
    return ifftn_to_tb(ifftn(kgrid_array, axes=np.arange(ndim)))


def ifftn_to_tb(ifft_array: np.ndarray) -> _tb_type:
    """
    Convert the result of `scipy.fft.ifftn` to a tight-binding dictionary.

    Parameters
    ----------
    ifft_array :
        Result of `scipy.fft.ifftn` to convert to a tight-binding dictionary.
        The input to `scipy.fft.ifftn` should be from `tb_to_khamvector`.
    Returns
    -------
    :
        Tight-binding dictionary.
    """
    size = ifft_array.shape[:-2]

    keys = [np.arange(-size[0] // 2 + 1, size[0] // 2) for i in range(len(size))]
    keys = itertools.product(*keys)
    return {tuple(k): ifft_array[tuple(k)] for k in keys}


def tb_to_kfunc(tb: _tb_type) -> Callable:
    """
    Fourier transforms a real-space tight-binding model to a k-space function.

    Parameters
    ----------
    tb :
        Tight-binding dictionary.

    Returns
    -------
    :
        A function that takes a k-space vector and returns a complex np.array.

    Notes
    -----
    Function doesn't work for zero dimensions.
    """

    def kfunc(k):
        ham = 0
        for vector in tb.keys():
            ham += tb[vector] * np.exp(-1j * np.dot(k, np.array(vector, dtype=float)))
        return ham

    return kfunc



from itertools import product
import numpy as np

from meanfi.tb.tb import _tb_type
from meanfi.mf import fermi_on_kgrid
from meanfi.tb.transforms import tb_to_kgrid


def guess_tb(
    tb_keys: list[tuple[None] | tuple[int, ...]], ndof: int, scale: float = 1
) -> _tb_type:
    """Generate hermitian guess tight-binding dictionary.

    Parameters
    ----------
    tb_keys :
       List of hopping vectors (tight-binding dictionary keys) the guess contains.
    ndof :
        Number internal degrees of freedom within the unit cell.
    scale :
        Scale of the random guess.
    Returns
    -------
    :
        Hermitian guess tight-binding dictionary.
    """
    guess = {}
    for vector in tb_keys:
        if vector not in guess.keys():
            amplitude = scale * np.random.rand(ndof, ndof)
            phase = 2 * np.pi * np.random.rand(ndof, ndof)
            rand_hermitian = amplitude * np.exp(1j * phase)
            if np.linalg.norm(np.array(vector)) == 0:
                rand_hermitian += rand_hermitian.T.conj()
                rand_hermitian /= 2
                guess[vector] = rand_hermitian
            else:
                guess[vector] = rand_hermitian
                guess[tuple(-np.array(vector))] = rand_hermitian.T.conj()

    return guess


def generate_tb_keys(cutoff: int, dim: int) -> list[tuple[None] | tuple[int, ...]]:
    """Generate tight-binding dictionary keys up to a cutoff.

    Parameters
    ----------
    cutoff :
        Maximum distance along each dimension to generate tight-bindign dictionary keys for.
    dim :
        Dimension of the tight-binding dictionary.

    Returns
    -------
    :
        List of generated tight-binding dictionary keys up to a cutoff.
    """
    return [*product(*([[*range(-cutoff, cutoff + 1)]] * dim))]


def fermi_energy(tb: _tb_type, filling: float, nk: int = 100):
    """
    Calculate the Fermi energy of a given tight-binding dictionary.

    Parameters
    ----------
    tb :
        Tight-binding dictionary.
    filling :
        Number of particles in a unit cell.
        Used to determine the Fermi level.
    nk :
        Number of k-points in a grid to sample the Brillouin zone along each dimension.
        If the system is 0-dimensional (finite), this parameter is ignored.

    Returns
    -------
    :
        Fermi energy.
    """
    kham = tb_to_kgrid(tb, nk)
    vals = np.linalg.eigvalsh(kham)
    return fermi_on_kgrid(vals, filling)



# %%
import numpy as np
import pytest

from meanfi.tb import utils
from meanfi.tb.tb import compare_dicts
from meanfi import Model, solver, guess_tb, add_tb, fermi_energy

# %%
repeat_number = 10


# %%
@pytest.mark.parametrize("seed", range(repeat_number))
def test_zero_hint(seed):
    """Test the zero interaction case for the tight-binding model."""
    np.random.seed(seed)

    cutoff = np.random.randint(1, 4)
    dim = np.random.randint(0, 3)
    ndof = np.random.randint(2, 10)
    filling = np.random.randint(1, ndof)
    random_hopping_vecs = utils.generate_tb_keys(cutoff, dim)

    zero_key = tuple([0] * dim)
    h_0_random = guess_tb(random_hopping_vecs, ndof, scale=1)
    h_int_only_phases = guess_tb(random_hopping_vecs, ndof, scale=0)
    guess = guess_tb(random_hopping_vecs, ndof, scale=1)
    model = Model(h_0_random, h_int_only_phases, filling=filling)

    mf_sol = solver(model, guess, nk=40, optimizer_kwargs={"M": 0, "f_tol": 1e-10})
    h_fermi = fermi_energy(mf_sol, filling=filling, nk=20)
    mf_sol[zero_key] -= h_fermi * np.eye(mf_sol[zero_key].shape[0])

    compare_dicts(add_tb(mf_sol, h_0_random), h_0_random, atol=1e-10)



# %%
import itertools as it

import numpy as np
import pytest
from scipy.fftpack import ifftn

from meanfi.tb.tb import compare_dicts
from meanfi.tb.utils import guess_tb, generate_tb_keys
from meanfi.tb.transforms import ifftn_to_tb, tb_to_kgrid, tb_to_kfunc

repeat_number = 10

# %%
ndim = 2
max_order = 5
matrix_size = 5
nk = 10


@pytest.mark.parametrize("seed", range(repeat_number))
def test_fourier(seed):
    """Test the Fourier transformation of the tight-binding model."""
    np.random.seed(seed)
    keys = [np.arange(-max_order + 1, max_order) for i in range(ndim)]
    keys = it.product(*keys)
    h_0 = {key: (np.random.rand(matrix_size, matrix_size) - 0.5) * 2 for key in keys}
    kham = tb_to_kgrid(h_0, nk=nk)
    tb_new = ifftn_to_tb(ifftn(kham, axes=np.arange(ndim)))
    compare_dicts(h_0, tb_new)


@pytest.mark.parametrize("seed", range(repeat_number))
def test_kfunc(seed):
    np.random.seed(seed)
    cutoff = np.random.randint(1, 4)
    dim = np.random.randint(1, 3)
    ndof = np.random.randint(2, 10)
    nk = np.random.randint(3, 10)
    random_hopping_vecs = generate_tb_keys(cutoff, dim)
    random_tb = guess_tb(random_hopping_vecs, ndof, scale=1)

    kfunc = tb_to_kfunc(random_tb)
    kham = tb_to_kgrid(random_tb, nk=nk)

    # evaluate kfunc on same grid as kham
    ks = np.linspace(-np.pi, np.pi, nk, endpoint=False)
    ks = np.concatenate((ks[nk // 2 :], ks[: nk // 2]), axis=0)
    k_pts = np.tile(ks, dim).reshape(dim, nk)

    ham_kfunc = []
    for k in it.product(*k_pts):
        ham_kfunc.append(kfunc(k))
    ham_kfunc = np.array(ham_kfunc).reshape(*kham.shape)

    assert np.allclose(kham, ham_kfunc)



# %%
import numpy as np
import pytest

from meanfi.kwant_helper import kwant_examples, utils
from meanfi import (
    Model,
    fermi_energy,
    solver,
    tb_to_kgrid,
    guess_tb,
    add_tb,
)


def compute_gap(tb, fermi_energy=0, nk=100):
    """Compute gap.

    Parameters
    ----------
    tb : dict
        Tight-binding model for which to compute the gap.
    fermi_energy : float
     Fermi energy.
    nk : int
     Number of k-points to sample along each dimension.

    Returns
    -------
     gap : float
     Indirect gap.
    """
    kham = tb_to_kgrid(tb, nk)
    vals = np.linalg.eigvalsh(kham)

    emax = np.max(vals[vals <= fermi_energy])
    emin = np.min(vals[vals > fermi_energy])
    return np.abs(emin - emax)


repeat_number = 5
# %%
graphene_builder, int_builder = kwant_examples.graphene_extended_hubbard()
h_0 = utils.builder_to_tb(graphene_builder)


# %%
def gap_prediction(U, V):
    """Test if the mean-field theory predicts the gap correctly for a given U and V.

    Parameters
    ----------
    U : float
        The Hubbard U parameter. Rounded to one decimal.
    V : float
        The nearest-neighbor interaction parameter. Rounded to one decimal.
    """
    U = np.round(U, 1)
    V = np.round(V, 1)
    params = {"U": U, "V": V}

    # Compare to phase diagram in https://arxiv.org/pdf/1204.4531.pdf
    upper_phase_line = 0.181 * U + 0.416
    lower_phase_line = 1.707 * U - 3.823
    triple_point = (2.78, 0.92)

    gapped = False
    if triple_point < (U, V):
        gapped = True
    elif (upper_phase_line < V) | (lower_phase_line > V):
        gapped = True

    # the mean-field calculation
    filling = 2
    nk = 40

    h_int = utils.builder_to_tb(int_builder, params)
    guess = guess_tb(frozenset(h_int), len(list(h_0.values())[0]))
    model = Model(h_0, h_int, filling)

    mf_sol = solver(model, guess, nk=nk, optimizer_kwargs={"M": 0, "f_tol": 1e-8})
    mf_full = add_tb(h_0, mf_sol)
    EF = fermi_energy(mf_full, filling=filling, nk=nk)
    gap = compute_gap(mf_full, fermi_energy=EF, nk=200)

    # Check if the gap is predicted correctly
    if gap > 0.1:
        gapped_predicted = True
    else:
        gapped_predicted = False
    assert (
        gapped == gapped_predicted
    ), f"Mean-field theory failed to predict the gap for U = {U}, V = {V}"


# %%
@pytest.mark.parametrize("seed", range(repeat_number))
def test_gap(seed):
    """Test if the mean-field theory predicts the gap correctly for a random U and V."""
    np.random.seed(seed)
    U = np.random.uniform(0, 4)
    V = np.random.uniform(0, 1)
    gap_prediction(U, V)



import numpy as np

from meanfi.tb.tb import _tb_type


def expectation_value(density_matrix: _tb_type, observable: _tb_type) -> complex:
    """Compute the expectation value of an observable with respect to a density matrix.

    Parameters
    ----------
    density_matrix :
        Density matrix tight-binding dictionary.
    observable :
        Observable tight-binding dictionary.

    Returns
    -------
    :
        Expectation value.
    """
    return np.sum(
        [
            np.trace(observable[k] @ density_matrix[tuple(-np.array(k))])
            for k in frozenset(density_matrix) & frozenset(observable)
        ]
    )



"Mean-field tight-binding solver"

try:
    from ._version import __version__, __version_tuple__
except ImportError:
    __version__ = "unknown"
    __version_tuple__ = (0, 0, "unknown", "unknown")

from .mf import (
    density_matrix,
    meanfield,
)
from .solvers import solver
from .model import Model
from .observables import expectation_value
from .tb.tb import add_tb, scale_tb
from .tb.transforms import tb_to_kgrid, kgrid_to_tb, tb_to_kfunc
from .tb.utils import guess_tb, fermi_energy


__all__ = [
    "solver",
    "Model",
    "expectation_value",
    "add_tb",
    "scale_tb",
    "guess_tb",
    "fermi_energy",
    "density_matrix",
    "meanfield",
    "tb_to_kgrid",
    "tb_to_kfunc",
    "kgrid_to_tb",
    "__version__",
    "__version_tuple__",
]



import numpy as np
import pytest
import kwant

import itertools as it

from meanfi.kwant_helper.utils import builder_to_tb, tb_to_builder
from meanfi.tb.utils import generate_tb_keys, guess_tb
from meanfi.tb.tb import compare_dicts

repeat_number = 3


@pytest.mark.parametrize("seed", range(repeat_number))
def test_kwant_conversion(seed):
    """Test conversion between Kwant and meanfi"""
    np.random.seed(seed)
    ndim = np.random.randint(1, 3)
    cutoff = np.random.randint(1, 3)
    sites_in_cell = np.random.randint(1, 4)
    ndof_per_site = [np.random.randint(1, 4) for site in range(sites_in_cell)]
    keyList = generate_tb_keys(cutoff, ndim)

    vecs = np.random.rand(ndim, ndim)

    # set a dummy lattice to read sites from
    lattice = kwant.lattice.general(
        vecs,
        basis=np.random.rand(sites_in_cell, ndim) @ vecs,
        norbs=ndof_per_site,
    )

    dummy_tb = kwant.Builder(kwant.TranslationalSymmetry(*lattice.prim_vecs))
    for i, sublattice in enumerate(lattice.sublattices):
        dummy_tb[lattice.shape(lambda pos: True, tuple(ndim * [0]))] = np.eye(
            ndof_per_site[i]
        )

    # generate random and generate builder from it
    random_tb = guess_tb(keyList, sum(ndof_per_site))
    random_builder = tb_to_builder(
        random_tb, list(dummy_tb.sites()), dummy_tb.symmetry.periods
    )
    # convert builder back to tb and compare
    random_builder_tb = builder_to_tb(random_builder)
    compare_dicts(random_tb, random_builder_tb)


@pytest.mark.parametrize("seed", range(repeat_number))
def test_kwant_supercell(seed):
    """Test with Kwant supercell and callable onsite and hoppings."""
    np.random.seed(seed)
    ndim = np.random.randint(1, 3)
    sites_in_cell = np.random.randint(1, 4)
    ndof_per_site = [np.random.randint(1, 4) for site in range(sites_in_cell)]
    n_cells = np.random.randint(1, 4)

    vecs = np.random.rand(ndim, ndim)

    # set a dummy lattice to read sites from
    lattice = kwant.lattice.general(
        vecs,
        basis=np.random.rand(sites_in_cell, ndim) @ vecs,
        norbs=ndof_per_site,
    )

    def random_matrix_kwant_digest(n, m, k):
        matrix = np.zeros((n, m))
        for i in zip(it.product(range(n), range(m))):
            matrix[i[0]] = kwant.digest.uniform(str(n * m * np.prod(i[0]) + k))
        return matrix

    def onsite(site, alpha, beta):
        n = site.family.norbs
        amplitude = alpha * random_matrix_kwant_digest(n, n, 0)
        phase = 1j * 2 * np.pi * beta * random_matrix_kwant_digest(n, n, 1)
        onsite_matrix = amplitude * phase
        onsite_matrix += onsite_matrix.conj().T
        return onsite_matrix

    def hopping(site1, site2, gamma, delta):
        n1 = site1.family.norbs
        n2 = site2.family.norbs
        amplitude = gamma * random_matrix_kwant_digest(n1, n2, 0)
        phase = 1j * 2 * np.pi * delta * random_matrix_kwant_digest(n1, n2, 1)
        hopping_matrix = amplitude * phase
        return hopping_matrix

    random_builder = kwant.Builder(
        kwant.TranslationalSymmetry(*n_cells * lattice.prim_vecs)
    )
    for i, sublattice in enumerate(lattice.sublattices):
        random_builder[lattice.shape(lambda pos: True, tuple(ndim * [0]))] = onsite
    random_builder[lattice.neighbors()] = hopping

    params_num = np.random.rand(4)
    params = dict(
        alpha=params_num[0],
        beta=params_num[1],
        gamma=params_num[2],
        delta=params_num[3],
    )

    random_tb, data = builder_to_tb(random_builder, params=params, return_data=True)
    random_builder_test = tb_to_builder(random_tb, data["sites"], data["periods"])
    for site_pair in zip(it.product(data["sites"], data["sites"])):
        site1, site2 = site_pair[0]
        if site1 == site2:
            assert np.isclose(
                random_builder[site1](
                    site=site1, alpha=params["alpha"], beta=params["beta"]
                ),
                random_builder_test[site1],
            ).all()
        else:
            try:
                assert np.isclose(
                    random_builder[site1, site2](
                        site1, site2, gamma=params["gamma"], delta=params["delta"]
                    ),
                    random_builder_test[site1, site2],
                ).all()
            except KeyError:
                continue
            except:
                raise



from itertools import product
from typing import Callable, Optional
import inspect

import numpy as np
from scipy.sparse import coo_array
import kwant
from kwant.builder import Site
import kwant.lattice
import kwant.builder


from meanfi.tb.tb import _tb_type


def builder_to_tb(
    builder: kwant.builder.Builder, params: dict = {}, return_data: bool = False
) -> _tb_type:
    """Construct a tight-binding dictionary from a `kwant.builder.Builder` system.

    Parameters
    ----------
    builder :
       system to convert to tight-binding dictionary.
    params :
        Dictionary of parameters to evaluate the builder on.
    return_data :
        Returns dictionary with sites and number of orbitals per site.

    Returns
    -------
    :
        Tight-binding dictionary that corresponds to the builder.
    :
        Data with sites and number of orbitals. Only if `return_data=True`.
    """
    prim_vecs = builder.symmetry.periods
    dims = len(prim_vecs)
    sites_list = [*builder.sites()]
    norbs_list = [site.family.norbs for site in builder.sites()]
    norbs_list = [1 if norbs is None else norbs for norbs in norbs_list]

    tb_norbs = sum(norbs_list)
    tb_shape = (tb_norbs, tb_norbs)
    onsite_idx = tuple([0] * dims)
    h_0 = {}

    for site, val in builder.site_value_pairs():
        site_idx = sites_list.index(site)
        tb_idx = np.sum(norbs_list[:site_idx]) + range(norbs_list[site_idx])
        row, col = np.array([*product(tb_idx, tb_idx)]).T

        if callable(val):
            param_keys = inspect.getfullargspec(val).args[1:]
            try:
                val = val(site, *[params[key] for key in param_keys])
            except KeyError as key:
                raise KeyError(f"Parameter {key} not found in params.")

        data = np.array(val).flatten()
        onsite_value = coo_array((data, (row, col)), shape=tb_shape).toarray()

        h_0[onsite_idx] = h_0.get(onsite_idx, 0) + onsite_value

    for (site1, site2), val in builder.hopping_value_pairs():
        site2_dom = builder.symmetry.which(site2)
        site2_fd = builder.symmetry.to_fd(site2)

        site1_idx, site2_idx = np.array(
            [sites_list.index(site1), sites_list.index(site2_fd)]
        )
        tb_idx1, tb_idx2 = [
            np.sum(norbs_list[:site1_idx]) + range(norbs_list[site1_idx]),
            np.sum(norbs_list[:site2_idx]) + range(norbs_list[site2_idx]),
        ]
        row, col = np.array([*product(tb_idx1, tb_idx2)]).T

        if callable(val):
            param_keys = inspect.getfullargspec(val).args[2:]
            try:
                val = val(site1, site2, *[params[key] for key in param_keys])
            except KeyError as key:
                raise KeyError(f"Parameter {key} not found in params.")

        data = np.array(val).flatten()
        hopping_value = coo_array((data, (row, col)), shape=tb_shape).toarray()

        hop_key = tuple(site2_dom)
        hop_key_back = tuple(-site2_dom)
        h_0[hop_key] = h_0.get(hop_key, 0) + hopping_value
        h_0[hop_key_back] = h_0.get(hop_key_back, 0) + hopping_value.T.conj()

    if return_data:
        data = {}
        data["periods"] = prim_vecs
        data["sites"] = sites_list
        return h_0, data
    else:
        return h_0


def tb_to_builder(
    h_0: _tb_type, sites_list: list[Site, ...], periods: np.ndarray
) -> kwant.builder.Builder:
    """
    Construct a `kwant.builder.Builder` from a tight-binding dictionary.

    Parameters
    ----------
    h_0 :
        Tight-binding dictionary.
    sites_list :
        List of sites in the builder's unit cell.
    periods :
        2d array with periods of the translational symmetry.

    Returns
    -------
    :
        `kwant.builder.Builder` that corresponds to the tight-binding dictionary.
    """

    builder = kwant.Builder(kwant.TranslationalSymmetry(*periods))
    onsite_idx = tuple([0] * len(list(h_0)[0]))

    norbs_list = [site.family.norbs for site in sites_list]
    norbs_list = [1 if norbs is None else norbs for norbs in norbs_list]

    def site_to_tbIdxs(site):
        site_idx = sites_list.index(site)
        return (np.sum(norbs_list[:site_idx]) + range(norbs_list[site_idx])).astype(int)

    # assemble the sites first
    for site in sites_list:
        tb_idxs = site_to_tbIdxs(site)
        value = h_0[onsite_idx][
            tb_idxs[0] : tb_idxs[-1] + 1, tb_idxs[0] : tb_idxs[-1] + 1
        ]
        builder[site] = value

    # connect hoppings within the unit-cell
    for site1, site2 in product(sites_list, sites_list):
        if site1 == site2:
            continue
        tb_idxs1 = site_to_tbIdxs(site1)
        tb_idxs2 = site_to_tbIdxs(site2)
        value = h_0[onsite_idx][
            tb_idxs1[0] : tb_idxs1[-1] + 1, tb_idxs2[0] : tb_idxs2[-1] + 1
        ]
        if np.all(value == 0):
            continue
        builder[(site1, site2)] = value

    # connect hoppings between unit-cells
    for key in h_0:
        if key == onsite_idx:
            continue
        for site1, site2_fd in product(sites_list, sites_list):
            site2 = builder.symmetry.act(key, site2_fd)
            tb_idxs1 = site_to_tbIdxs(site1)
            tb_idxs2 = site_to_tbIdxs(site2_fd)
            value = h_0[key][
                tb_idxs1[0] : tb_idxs1[-1] + 1, tb_idxs2[0] : tb_idxs2[-1] + 1
            ]
            if np.all(value == 0):
                continue
            builder[(site1, site2)] = value
    return builder


def build_interacting_syst(
    builder: kwant.builder.Builder,
    lattice: kwant.lattice.Polyatomic,
    func_onsite: Callable,
    func_hop: Optional[Callable] = None,
    max_neighbor: int = 1,
) -> kwant.builder.Builder:
    """
    Construct an auxiliary `kwant` system that encodes the interactions.

    Parameters
    ----------
    builder :
        Non-interacting `kwant.builder.Builder` system.
    lattice :
        Lattice of the system.
    func_onsite :
        Onsite interactions function.
    func_hop :
        Hopping/inter unit cell interactions function.
    max_neighbor :
        The maximal number of neighbouring unit cells (along a lattice vector)
        connected by interaction. Interaction goes to zero after this distance.

    Returns
    -------
    :
        Auxiliary `kwant.builder.Builder` that encodes the interactions of the system.
    """
    int_builder = kwant.builder.Builder(
        kwant.lattice.TranslationalSymmetry(*builder.symmetry.periods)
    )
    int_builder[builder.sites()] = func_onsite
    if func_hop is not None:
        for neighbors in range(max_neighbor):
            int_builder[lattice.neighbors(neighbors + 1)] = func_hop
    return int_builder



import numpy as np
import pytest
from scipy.optimize._nonlin import NoConvergence

from meanfi import (
    Model,
    guess_tb,
    add_tb,
    density_matrix,
)
from meanfi.solvers import solver_density, solver_mf
from meanfi.tb.utils import generate_tb_keys
from meanfi.tb.tb import compare_dicts

repeat_number = 5
ndim = 1
ndof = 4
cutoff = 2
U0 = 1
nk = 500
filling = ndof / 2


@pytest.mark.parametrize("seed", range(repeat_number))
def test_solver_consistency(seed):
    np.random.seed(seed)
    keys = generate_tb_keys(cutoff, ndim)
    h0 = guess_tb(keys, ndof)
    h_int = guess_tb(keys, ndof)
    h_int[keys[len(keys) // 2]] += U0
    guess = guess_tb(frozenset(h_int), ndof)
    _model = Model(h0, h_int, filling=filling)

    def solve_for_rho(model, solver):
        mf_sol = solver(
            model,
            mf_guess=guess,
            nk=nk,
            optimizer_kwargs={"M": 0, "f_tol": 1e-3, "maxiter": 10000},
        )
        mf_full = add_tb(h0, mf_sol)
        rho, _ = density_matrix(mf_full, filling=filling, nk=nk)
        rho = {key: rho[key] for key in h_int}
        return rho

    try:
        rho_dens = solve_for_rho(_model, solver_density)
        rho_mf = solve_for_rho(_model, solver_mf)
    except NoConvergence:
        guess = guess_tb(frozenset(h_int), ndof)
        rho_dens = solve_for_rho(_model, solver_density)
        rho_mf = solve_for_rho(_model, solver_mf)

    compare_dicts(rho_dens, rho_mf, atol=1e-2)



import numpy as np
from typing import Tuple

from meanfi.tb.tb import add_tb, _tb_type
from meanfi.tb.transforms import tb_to_kgrid, kgrid_to_tb


def density_matrix_kgrid(kham: np.ndarray, filling: float) -> Tuple[np.ndarray, float]:
    """Calculate density matrix on a k-space grid.

    Parameters
    ----------
    kham :
        Hamiltonian from which to construct the density matrix.
        The hamiltonian is sampled on a grid of k-points and has shape (nk, nk, ..., ndof, ndof),
        where ndof is number of internal degrees of freedom.
    filling :
        Number of particles in a unit cell.
        Used to determine the Fermi level.

    Returns
    -------
    :
        Density matrix on a k-space grid with shape (nk, nk, ..., ndof, ndof) and Fermi energy.
    """
    vals, vecs = np.linalg.eigh(kham)
    fermi = fermi_on_kgrid(vals, filling)
    unocc_vals = vals > fermi
    occ_vecs = vecs
    np.moveaxis(occ_vecs, -1, -2)[unocc_vals, :] = 0
    _density_matrix_krid = occ_vecs @ np.moveaxis(occ_vecs, -1, -2).conj()
    return _density_matrix_krid, fermi


def density_matrix(h: _tb_type, filling: float, nk: int) -> Tuple[_tb_type, float]:
    """Compute the real-space density matrix tight-binding dictionary.

    Parameters
    ----------
    h :
        Hamiltonian tight-binding dictionary from which to construct the density matrix.
    filling :
        Number of particles in a unit cell.
        Used to determine the Fermi level.
    nk :
        Number of k-points in a grid to sample the Brillouin zone along each dimension.
        If the system is 0-dimensional (finite), this parameter is ignored.

    Returns
    -------
    :
        Density matrix tight-binding dictionary and Fermi energy.
    """
    ndim = len(list(h)[0])
    if ndim > 0:
        kham = tb_to_kgrid(h, nk=nk)
        _density_matrix_krid, fermi = density_matrix_kgrid(kham, filling)
        return (
            kgrid_to_tb(_density_matrix_krid),
            fermi,
        )
    else:
        _density_matrix, fermi = density_matrix_kgrid(h[()], filling)
        return {(): _density_matrix}, fermi


def meanfield(density_matrix: _tb_type, h_int: _tb_type) -> _tb_type:
    """Compute the mean-field correction from the density matrix.

    Parameters
    ----------
    density_matrix :
        Density matrix tight-binding dictionary.
    h_int :
        Interaction hermitian Hamiltonian tight-binding dictionary.
    Returns
    -------
    :
        Mean-field correction tight-binding dictionary.

    Notes
    -----

    The interaction h_int must be of density-density type.
    For example, h_int[(1,)][i, j] = V means a repulsive interaction
    of strength V between two particles with internal degrees of freedom i and j
    separated by 1 lattice vector.
    """
    n = len(list(density_matrix)[0])
    local_key = tuple(np.zeros((n,), dtype=int))

    direct = {
        local_key: np.sum(
            np.array(
                [
                    np.diag(
                        np.einsum("pp,pn->n", density_matrix[local_key], h_int[vec])
                    )
                    for vec in frozenset(h_int)
                ]
            ),
            axis=0,
        )
    }

    exchange = {
        vec: -1 * h_int.get(vec, 0) * density_matrix[vec] for vec in frozenset(h_int)
    }
    return add_tb(direct, exchange)


def fermi_on_kgrid(vals: np.ndarray, filling: float) -> float:
    """Compute the Fermi energy on a grid of k-points.

    Parameters
    ----------
    vals :
        Eigenvalues of a hamiltonian sampled on a k-point grid with shape (nk, nk, ..., ndof, ndof),
        where ndof is number of internal degrees of freedom.
    filling :
        Number of particles in a unit cell.
        Used to determine the Fermi level.
    Returns
    -------
    :
        Fermi energy
    """
    norbs = vals.shape[-1]
    vals_flat = np.sort(vals.flatten())
    ne = len(vals_flat)
    ifermi = int(round(ne * filling / norbs))
    if ifermi >= ne:
        return vals_flat[-1]
    elif ifermi == 0:
        return vals_flat[0]
    else:
        fermi = (vals_flat[ifermi - 1] + vals_flat[ifermi]) / 2
        return fermi



import numpy as np

_tb_type = dict[tuple[int, ...], np.ndarray]


def add_tb(tb1: _tb_type, tb2: _tb_type) -> _tb_type:
    """Add up two tight-binding dictionaries together.

    Parameters
    ----------
    tb1 :
        Tight-binding dictionary.
    tb2 :
        Tight-binding dictionary.

    Returns
    -------
    :
        Sum of the two tight-binding dictionaries.
    """
    return {k: tb1.get(k, 0) + tb2.get(k, 0) for k in frozenset(tb1) | frozenset(tb2)}


def scale_tb(tb: _tb_type, scale: float) -> _tb_type:
    """Scale a tight-binding dictionary by a constant.

    Parameters
    ----------
    tb :
        Tight-binding dictionary.
    scale :
        Constant to scale the tight-binding dictionary by.

    Returns
    -------
    :
        Scaled tight-binding dictionary.
    """
    return {k: tb.get(k, 0) * scale for k in frozenset(tb)}


def compare_dicts(dict1: dict, dict2: dict, atol: float = 1e-10) -> None:
    """Compare two dictionaries."""
    for key in frozenset(dict1) | frozenset(dict2):
        assert np.allclose(dict1[key], dict2[key], atol=atol)
